/*  algebra.c  implements subalgebras of the Steenrod algebra */

#include "algebra.h"
#include <stdio.h>
#include "seqno.h"
#include "stop.h"

algebra thealgebra;

/* substitute for ~, gcc does some strange (= sick) things */
uint16 bitneg16(uint16 n) {  
  uint16 res=0; res = ~res; res ^= n; 
  return res;
}

int algebra_maxdegree(algebra *a) {
  int res = 0,n;
  if (!a->is_restricted) return 99999;
  for (n=0;n<15;n++) res += bitneg16(a->badbits[n]);
  return res;
}

void set_maxdegree(void) {
  MAXDEGREE = (1 << NALG) - 1;
  if (algebra_maxdegree(&thealgebra) < MAXDEGREE) MAXDEGREE = 99999;
}

void check_algebra(void) {
  int n;
  for (n=0;n<15;n++) {
    uint16 a = bitneg16(thealgebra.badbits[n]);
    if (0 != (a & (a+1))) error("check_algebra(..)");  
  }
}

void set_algebra(void) { /* should be called after every modification */
  check_algebra();
  set_maxdegree();
  thealgebra.topdeg = algebra_maxdegree(&thealgebra);
  if (thealgebra.topdeg < MAXDEGREE) {
    init_dimtab(thealgebra.topdeg);
  } else init_dimtab(MAXDEGREE);
}

void make_zero_algebra(algebra *a) {
  int n;
  a->is_restricted = 1;
  for (n=15;n--;) a->badbits[n] = ~0;
}

void allow_P(algebra *a, int i, int j) { /* allows P_{i+1}^j */
  a->badbits[i] &= bitneg16(1 << j);
}

void copy_algebra(algebra *a, algebra *b) {
  int n;
  a->is_restricted = b->is_restricted;
  for (n=15;n--;) a->badbits[n] = b->badbits[n];
}

int algebras_are_equal(algebra *a, algebra *b) {
  int n;
  if (a->is_restricted != b->is_restricted) return 0;
  for (n=15;n--;) if (a->badbits[n] != b->badbits[n]) return 0;
  return 1;
}

int algebra_is_alg_n(algebra *a) {
  uint16 aux; int n;
  if (!a->is_restricted) return -1;
  aux = bitneg16(a->badbits[0]);
  for (n=1;n<15;n++) 
    if (bitneg16(aux >>= 1) != (a->badbits[n])) return -1;
  aux = bitneg16(a->badbits[0]); n = 0;
  while ((aux >>= 1)) n++;
  return n;
}

int nb(uint16 n) {  
  int res=0; while (n) { res++, n >>= 1; }; return res;
}

char *algebra_name(algebra *a) {
  static char res[50];
  int n,m; char c,*d;
  if (!a->is_restricted) {
    sprintf(res,"A"); return res;
  }
  n = algebra_is_alg_n(a);
  if (n >= 0) {
    sprintf(res,"A(%d)",n); return res;
  }
  for (m=14;m>=0;m--) if (0 != bitneg16(a->badbits[m])) break;
  if (m < 0) error("algebra_name(..) internal error");
  for (n=0,c=' ',d=res;n<=m;n++,c='|') {
    sprintf(d,"%c%d",c,nb(bitneg16(a->badbits[n])));
    while (*++d) ;
  }
  return res+1;
}

char *algebra_path_name(algebra *a) {
  static char res[50];
  int n,m; char *d;
  if (!a->is_restricted) {
    sprintf(res,"res.A"); return res;
  }
  n = algebra_is_alg_n(a);
  if (n >= 0) {
    sprintf(res,"res.A%d",n); return res;
  }
  for (m=14;m>=0;m--) if (0 != bitneg16(a->badbits[m])) break;
  if (m < 0) error("algebra_path_name(..) internal error");
  sprintf(res,"res");
  for (n=0,d=res+3;n<=m;n++) {
    sprintf(d,".%d",nb(bitneg16(a->badbits[n])));
    while (*++d) ;
  }
  return res;
}

void clear_algebra(algebra *a) {
  int n;
  a->is_restricted = 0;
  for (n=15;n--;) a->badbits[n] = 0;
}

void xset_algebra_An(algebra *a, int n) { 
  int i,j;
  make_zero_algebra(a);
  for (i=0;i<=n;i++)
    for (j=0;i+j<=n;j++)
      allow_P(a,i,j);
  a->is_restricted = 1;
}

void set_algebra_An(int n) { 
  xset_algebra_An(&thealgebra,n); 
  set_algebra();
}

void set_algebra_Ainfty(void) { 
  clear_algebra(&thealgebra); 
  set_algebra();
}
