/* modules.h   -     elements, modules, etc.  */

#if !defined MODULES_DEF
#define MODULES_DEF

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "error.h"
#include "stoputils.h"
#include "linalg.h"

#define MAXELEMENT  MAXDIM /* maximal number of summands */
#define MAXGENS     3000  /* maximal number of gens per module */

typedef struct {
  int num;
  stop dat[MAXELEMENT];
} element;

typedef struct {
  int num;
  stop dat[MAXDIM];
} small_element;

typedef struct {
  int  num;         /* number of summands */
  stop *dat;        /* array of stops     */
} stoplist;

typedef struct {
  stoplist diff;    /* differential    */
  int deg;          /* internal degree */   
  int num;          /* number of summands */
} gen;

typedef union {
  char x[256];  /* makes module_header have exactly 256 bytes, so future */
                /* extensions in the 'data' component are possible.      */ 
  struct { 
    int mpn;         /* my phone number = 26219. recognizes corrupted files */
    int version;     /* version number of the data format used.             */
    time_t cr_date;  /* date of creation                                    */
    time_t up_date;  /* date of last modification                           */
    int sdeg;        /* file describes the module C_sdeg                    */
    int tmax;        /* file is complete through internal degree tmax       */
    algebra alg;     /* file refers to the algebra alg                      */
	 char fnme[50];   /* the file name                                       */
  } data;
} module_header;

typedef struct {
  module_header header;
  int sdeg; 
  int numgens;
  gen gens[MAXGENS];
} module;

typedef int (*headfunc)(module *);
typedef int (*genfunc)(module *, int);
typedef int (*stopfunc)(module *, int, stop *);

/* low level access to module files via parse_modulefile(...) */
/*
  what  parse_modulefile(s,m,hf,gf,sf,ef,mode)  will do:

    1) open the file Cs.  (if (mode & MOD_NOCREATE) Cs has to exist)
    2) read header, and call hf(m) if (hf != NULL)
    3) if (mode != MOD_HEADER) 
          4) for each gen g  call gf(m,numgen) if (gf != NULL)
	  5) read dg from file (into memory unless MOD_NOLOAD requested)
	  6) if (sf != NULL) for each stop x in dg call sf(m,numgen,&x)
	  7) if (ef != NULL) call ef(m,numgen)  (ef = end function)
    8) close file Cs

  parse_modulefile will exit, if one of hf, gf, sf or ef returns 0 
*/

#define MOD_NORMAL   0
#define MOD_NOLOAD   1
#define MOD_HEADER   2
#define MOD_NOCREATE 4

int  parse_modulefile(int s, module *m, 
		      headfunc hf,
		      genfunc  gf,
		      stopfunc sf,
		      genfunc  ef,
		      int mode);

/* parse_datafile(...) is equivalent to parse_modulefile(...), except
	that it allows you to choose an arbitrary filename for the module */
 
int  parse_datafile(char *fnme, int s, module *m, 
		      headfunc hf,
		      genfunc  gf,
		      stopfunc sf,
		      genfunc  ef,
		      int mode);


/* the following routines are interfaces to parse_modulefile */
int read_module_no_diffs(module *m, int s);
int read_module(module *m, int s);

int read_sq0file(module *m, int s);
int read_hurfile(module *m, int s, int n);

/* routines to fake initial modules */
void make_zero_module(module *m);
void make_C0_module(module *m);

void fake_doubl0(module *m);
void fake_hur0(module *m);

/* write dg to file *fp */
void write_gen(FILE *fp, gen *g);

/* dispose_module(m) frees the memory allocated for m */
void dispose_module(module *m);

void update_header(module_header *mh);
int  read_header(FILE *fp, module_header *mh);
void print_header(module_header *mh);

void print_element(element *e);

#define DUMP_ASCII 0 /* the only type currently supported */
#define DUMP_TEX   1 /* TeX mode */
void dump_module(FILE *fp, module *m, int type);

int num_stop(element *e, stop *s);
int num_stop_same_signature(element *e, stop *s);
void add_stop_to_element(element *e, stop *s);

void modulebasis(element *b, module *m, int t, signature *sig);
int countmodulebasis(module *m, int t, int max); /* just return dimension */

/* call init_stoplist_buffer(); before use */
/* (allthough it doesnt do anything yet)  */
void init_stoplist_buffer(void);

int  malloc_stoplist(stoplist *s, int num);
void print_stoplist(stoplist *x);
void dispose_stoplist(stoplist *s);

void gen_from_element(gen *g, element *e, int dg);

#endif
