/* optimized multiplication routine */

#include "optmult.h"

/*
   rationale:

      The other multiplication routines are optimized for the task
      of computing the product  ab  for fixed Milnor basis elements
      a  and  b.  In practice, however, one encouters the computation 
      of

               a ( b1 + b2 + b3 + ... + bN ),  N large

      and the optimization for this task is different.

      The routine

         optimized_multiply
            (stop *s, stoplist *diff, unit *(dat[]), modinfo *mi);

      computes  s * diff  and adds the result to the vectors dat[k] != NULL.

   plan:

      1)    
         Compute max(b1,b2,...,bN)  (taken entrywise)
         This gives restrictions for the entries in the admissible
         matrices.

         One day the maximum should be computed 
            a)  for each gen that is read from the datafile
            b)  for each gen that is newly introduced
         ie. max should be part of the internal representation of
         the resolution.

      2)    
         Compute according to the following scheme:

            for (admissible matrices for a) 
               go through all b's and compute the result
               add this result to dat.

*/

typedef struct {
  short int dat[NALG+1][NALG+1];
  short int rem[NALG+1][NALG+1];
  short int msk[NALG+1][NALG+1];
  short int sum[NALG+1][NALG+1];
  unit **vec;             /* pointers to vectors for the result */
  modinfo *mi;            /* module info, needed to access vec  */
  stoplist *diff;         /* the second factor                  */
  stop *first;
  int deg;
  element *basis;         /* for make_matrix_same_sig          */
  stop *sig;              /* fixed signature case              */
  unit *vecbas;           /* vector for the basis based computation */
}  optmatrix_type;

optmatrix_type OM;

void make_max_stop(stop *res, stoplist *diff) {
  stop *s = diff->dat; int n = diff->num; 
  clear_stop(res);
  while (n--) {
    int i;
    for (i=NALG;i--;) {
      int x = s->dat[i];
      if (x > res->dat[i]) res->dat[i] = x; 
    }
    s++;
  }
}

void init_optmatrix(stop *first, stop *max, unit *(dat[]), modinfo *mi, 
               stoplist *diff, int deg) {
   int n;
   for (n=NALG;n--;) {
      OM.msk[NALG][n] = 0; /* no initial restrictions */
      OM.sum[NALG][n+1] = 0;
      OM.rem[NALG][n+1] = max->dat[n];
      OM.dat[NALG][n] = 0;
   }
   OM.vec  = dat;
   OM.mi   = mi;
   OM.diff = diff;
   OM.first = first;
	OM.deg = deg;
}

int opt_first_row(int n, int deg) {
   /* assume that rows n+1,n+2,...,NALG-1 have been filled, so we */
   /* got restrictions in rem[n] and msk[n]                       */
   unsigned char c=NALG; 
   while (c--) {
      short int x = deg >> c;
      if (c) {
         short int badbits = OM.msk[n][c];
         if (x > OM.rem[n][c]) x = OM.rem[n][c];
	 REMOVE_BADBITS(x,badbits);
         OM.msk[n-1][c+1] = badbits | x;
         OM.rem[n-1][c] = OM.rem[n][c] - x;
         OM.sum[n-1][c] = OM.sum[n][c] + x;
      } else OM.msk[n-1][1] = x;
      OM.dat[n][c] = x;
      deg -= x << c;
   }
   if (0 != deg) return 0;
   return 1;
}

int opt_next_row(int n) {
   short int c=1,deg=OM.dat[n][0],x;
   do {
      while ((c < NALG) && (0 == OM.dat[n][c])) c++;
      if (c == NALG) return 0;
      x = OM.dat[n][c]-1;
      if (c) {
         short int badbits = OM.msk[n][c];
	 REMOVE_BADBITS(x,badbits);
         OM.msk[n-1][c+1] = badbits | x;
         OM.rem[n-1][c] = OM.rem[n][c] - x;
         OM.sum[n-1][c] = OM.sum[n][c] + x;
      } else OM.msk[n-1][c+1] = x;
      deg += (OM.dat[n][c] - x) << c;
      OM.dat[n][c] = x; 
      while (c--) {
         x = deg >> c;
         if (c) {
            short int badbits = OM.msk[n][c];
            if (x > OM.rem[n][c]) x = OM.rem[n][c];
	    REMOVE_BADBITS(x,badbits);
            OM.msk[n-1][c+1] = badbits | x;
            OM.rem[n-1][c] = OM.rem[n][c] - x;
            OM.sum[n-1][c] = OM.sum[n][c] + x;
         } else OM.msk[n-1][c+1] = x;
         OM.dat[n][c] = x;
         deg -= x << c;
      }
   } while (0 != deg);
   return 1;
}

void opt_handle_row(int n) {
   if (n) {
      if (opt_first_row(n,OM.first->dat[n-1])) 
         do {
            opt_handle_row(n-1);
         } while (opt_next_row(n));
   } else {   /* get result */
      stop *s = OM.diff->dat; int n = OM.diff->num;
      while (n--) {
         unsigned char c,nok;
         stop res;
         for (nok=c=0;c < NALG;c++) {
            short int x = (s->dat[c]);
            if (nok |= (x < OM.sum[0][c+1])) break;
            x -= OM.sum[0][c+1];
            if (nok |= (x & OM.msk[0][c+1])) break;
            res.dat[c] = x | OM.msk[0][c+1];
         }
         if (!nok) {
            int k; unit z = 1; unit **aux = OM.vec;
            res.id = s->id;
            k = module_seqno_with_degree(&res,OM.mi,OM.deg);
            z <<= (k & UNMASK); k >>= UNSIZE;
            while (NULL != *aux) {
               (*aux)[k] ^= z;
               aux++;
            }
         }
         s++;
      }
   }
}

void optimized_multiply(stop *s, stoplist *diff, unit *(dat[]), 
								modinfo *mi, int deg) {
   stop max;
   make_max_stop(&max,diff);
   init_optmatrix(s,&max,dat,mi,diff,deg);
   opt_handle_row(NALG);
}

/**************************************************************************

   same optimization for make_matrix_same_sig

**************************************************************************/

void init_optmatrix_basis(stop *first, stop *max, unit *dat, element *basis, 
               stoplist *diff, stop *sig) {
   int n;
   for (n=NALG;n--;) {
      OM.msk[NALG][n] = 0; /* no initial restrictions */
      OM.sum[NALG][n+1] = 0;
      OM.rem[NALG][n+1] = max->dat[n];
      OM.dat[NALG][n] = 0;
   }
   OM.vecbas = dat;
   OM.basis  = basis;
   OM.diff   = diff;
   OM.first  = first;
   OM.sig    = sig;
}

void opt_handle_row_basis(int n) {
  if (n) {
    if (opt_first_row(n,OM.first->dat[n-1])) 
      do {
	opt_handle_row_basis(n-1);
      } while (opt_next_row(n));
  } else {   /* get result */
    stop *s = OM.diff->dat; int n = OM.diff->num;
    while (n--) {
      unsigned char c,nok;
      stop res;
      for (nok=c=0;c < NALG;c++) {
	short int x = s->dat[c];
	if (nok |= (x < OM.sum[0][c+1])) break;
	x -= OM.sum[0][c+1];
	if (nok |= (x & OM.msk[0][c+1])) break;
	res.dat[c] = x | OM.msk[0][c+1];
      }
      if (!nok) {
	int k; unit z = 1;
	res.id = s->id;       
	set_signature(&res,OM.sig);
	/*fprintf_stop(stdout,&res);*/
	k = num_stop_same_signature(OM.basis,&res);
	if (-1 == k) {
            FILE *fp;
            if (NULL != (fp = fopen("error.report","w"))) {
                fprintf(fp,"\nstop "); 
                fprintf_stop(fp,&res);
                fprintf(fp," not found\n"); 
                fprintf(fp,"signature = ");
                fprintf_stop(fp,OM.sig);
                fprintf(fp,"\n");
                fprintf(fp,"basis = ");
                fprint_element(fp,OM.basis);
                fclose(fp);
            }
	  error("opt_make_matrix_same_sig");
	} else {
	  z <<= (k & UNMASK); 
	  (OM.vecbas)[k >> UNSIZE] ^= z;
	}
      }
      s++;
    }
  }
}

void optimized_multiply_basis(stop *s, stoplist *diff, unit *dat, 
                     element *basis, stop *sig) {
   stop max,aux; int n;
/*
printf("opt_mult_basis: ");
fprintf_stop(stdout,s);
printf(" * ");
print_stoplist(diff);
printf(" = ");
*/
   copy_stop(&aux,s);
   for (n=0;n<NALG;n++) aux.dat[n] |= theprofile.mask[n];
   make_max_stop(&max,diff);
   init_optmatrix_basis(&aux,&max,dat,basis,diff,sig);
   opt_handle_row_basis(NALG);
}

void opt_make_matrix_same_sig(
   module *ms, module *mt, int t, signature *sig,
   small_element *bs, small_element *bt, matrix *mat) {
   int n; 
   stoplist *diff; int numgen;
   /* first make bases */
   modulebasis((element *) bs,ms,t,sig);
   modulebasis((element *) bt,mt,t,sig);
   set_values(mat,bs->num,bt->num);
   clear_matrix(mat,bs->num,bt->num);
   for (n=0;n<bs->num;n++) {
      numgen = bs->dat[n].id; diff = &(ms->gens[numgen].diff);
      
      optimized_multiply_basis(&(bs->dat[n]),diff,mat->dat[n],(element *) bt,sig);
   
   }
}


